// OE2.cpp - OutlookExpress-To - OE5/6 Multi Converter
// Copyright (C) 2002 Tietew
//
#include "pch.h"
#include "OE2.h"
#include "utils.h"
#include "plugin.h"

HINSTANCE g_hinst;
volatile bool g_abort;
volatile bool g_pause;
static UINT disable_ids[] = {
	IDC_START, IDC_SOURCE, IDC_BROWSE_SRC, IDC_PLUGIN,
	IDC_DEST, IDC_BROWSE_DEST
};

int LoadPlugins(HWND hwnd)
{
	int n = 0;
	WIN32_FIND_DATA fd;
	HANDLE fh = FindFirstFile("*.dll", &fd);
	if(fh == INVALID_HANDLE_VALUE)
		return 0;
	do {
		HMODULE module = LoadLibrary(fd.cFileName);
		if(module != NULL) {
			Plugin *p = new Plugin;
			if(p->Load(module)) {
				char *name = p->GetPluginName();
				int i = (int)SendMessage(hwnd, CB_ADDSTRING, 0, (LPARAM)name);
				SendMessage(hwnd, CB_SETITEMDATA, i, (LPARAM)p);
				free(name);
				++n;
			} else
				delete p;
		}
	} while(FindNextFile(fh, &fd));
	FindClose(fh);
	
	return n;
}

void LoadSources(HWND hwnd)
{
	LONG r;
	HKEY rkey, hkey, hkey2;
	DWORD i, type, size;
	char name[1024];
	char *username, *store;
	
	r = RegOpenKeyEx(HKEY_CURRENT_USER, "Identities", 0, KEY_READ, &rkey);
	if(r != ERROR_SUCCESS)
		return;
	
	for(i = 0; ; ++i) {
		size = sizeof name;
		if(RegEnumKeyEx(rkey, i++, name, &size, 0, 0, 0, 0) != ERROR_SUCCESS)
			break;
		if(RegOpenKeyEx(rkey, name, 0, KEY_READ, &hkey) != ERROR_SUCCESS)
			continue;
		do {
			r = RegQueryValueEx(hkey, "UserName", 0, &type, 0, &size);
			if(r != ERROR_SUCCESS || type != REG_SZ)
				break;
			username = (char *)_alloca(++size);
			r = RegQueryValueEx(hkey, "UserName", 0, &type, (LPBYTE)username, &size);
			if(r != ERROR_SUCCESS)
				break;
			
			r = RegOpenKeyEx(hkey, "Software\\Microsoft\\Outlook Express\\5.0", 0, KEY_READ, &hkey2);
			if(r != ERROR_SUCCESS)
				break;
			do {
				r = RegQueryValueEx(hkey2, "Store Root", 0, &type, 0, &size);
				if(r != ERROR_SUCCESS || (type != REG_SZ && type != REG_EXPAND_SZ))
					break;
				store = (char *)_alloca(++size);
				r = RegQueryValueEx(hkey2, "Store Root", 0, &type, (LPBYTE)store, &size);
				if(r != ERROR_SUCCESS)
					break;
				
				size *= 5;
				if(size < MAX_PATH * 2)
					size = MAX_PATH * 2;
				LPSTR path = (LPSTR)_alloca(size);
				if(type == REG_EXPAND_SZ) {
					if(!ExpandEnvironmentStrings(store, path, size))
						path = NULL;
				} else
					path = store;
				
				if(path) {
					int i = (int)SendMessage(hwnd, CB_ADDSTRING, 0, (LPARAM)username);
					SendMessage(hwnd, CB_SETITEMDATA, i, (LPARAM)strdup(path));
				}
			} while(0);
			RegCloseKey(hkey2);
		} while(0);
		RegCloseKey(hkey);
	}
	RegCloseKey(rkey);
}

BOOL OnInitDialog(HWND hwnd)
{
	if(LoadPlugins(GetDlgItem(hwnd, IDC_PLUGIN)) == 0) {
		MyMessageBox(NULL, IDS_NO_PLUGIN, MB_ICONEXCLAMATION | MB_OK);
		EndDialog(hwnd, IDCANCEL);
		return TRUE;
	}
	
	HWND cb = GetDlgItem(hwnd, IDC_SOURCE);
	LoadSources(cb);
	SendMessage(cb, CB_SETCURSEL, 0, 0);
	
	char mesg[128];
	LoadString(g_hinst, IDS_SEL_PLUGIN, mesg, countof(mesg));
	SetDlgItemText(hwnd, IDC_NOTE, mesg);
	
	EnableDlgItem(hwnd, IDC_SETTING, false);
	EnableDlgItem(hwnd, IDC_START, false);
	EnableDlgItem(hwnd, IDC_ABORT, false);
	SendDlgItemMessage(hwnd, IDC_PROGRESS1, PBM_SETRANGE, 0, MAKELPARAM(0, 100));
	
	HICON icon;
	icon = LoadIcon(g_hinst, MAKEINTRESOURCE(IDR_MAINFRAME));
	SendMessage(hwnd, WM_SETICON, ICON_BIG, (LPARAM)icon);
	icon = (HICON)LoadImage(g_hinst, MAKEINTRESOURCE(IDR_MAINFRAME),
		IMAGE_ICON, 16, 16, 0);
	SendMessage(hwnd, WM_SETICON, ICON_SMALL, (LPARAM)icon);
	
	return TRUE;
}

BOOL OnDestroy(HWND hwnd)
{
	HWND cb = GetDlgItem(hwnd, IDC_SOURCE);
	int len = (int)SendMessage(cb, CB_GETCOUNT, 0, 0);
	for(int i = 0; i < len; ++i) {
		char *name = (char *)SendMessage(cb, CB_GETITEMDATA, i, 0);
		free(name);
	}
	cb = GetDlgItem(hwnd, IDC_PLUGIN);
	len = (int)SendMessage(cb, CB_GETCOUNT, 0, 0);
	for(int i = 0; i < len; ++i) {
		Plugin *p = (Plugin *)SendMessage(cb, CB_GETITEMDATA, i, 0);
		delete p;
	}
	return FALSE;
}

int CALLBACK BrowseCallback(HWND hwnd, UINT msg, LPARAM lp, LPARAM context)
{
	if(msg == BFFM_INITIALIZED)
		SendMessage(hwnd, BFFM_SETSELECTION, TRUE, context);
	return 0;
}

char *BrowseFolder(HWND hwnd, const char *path, UINT titleid)
{
	char disp[MAX_PATH];
	char title[128];
	
	BROWSEINFO bi;
	bi.hwndOwner = hwnd;
	bi.pidlRoot = NULL;
	bi.pszDisplayName = disp;
	LoadString(g_hinst, titleid, title, countof(title));
	bi.lpszTitle = title;
	bi.ulFlags = BIF_EDITBOX | BIF_NEWDIALOGSTYLE | BIF_RETURNONLYFSDIRS;
	bi.lpfn = BrowseCallback;
	bi.lParam = (LPARAM)path;
	bi.iImage = 0;
	
	LPITEMIDLIST pidl = SHBrowseForFolder(&bi);
	if(pidl == NULL)
		return NULL;
	SHGetPathFromIDList(pidl, disp);
	
	IMalloc *p;
	SHGetMalloc(&p);
	p->Free(pidl);
	p->Release();
	
	return strdup(disp);
}

BOOL OnBrowseSrc(HWND hwnd)
{
	char *path1, *path2;
	HWND cb = GetDlgItem(hwnd, IDC_SOURCE);
	int i = (int)SendMessage(cb, CB_GETCURSEL, 0, 0);
	
	if(i == CB_ERR) {
		int len = GetWindowTextLength(cb);
		path1 = (char *)malloc(len + 1);
		GetWindowText(cb, path1, len + 1);
	} else
		path1 = (char *)SendMessage(cb, CB_GETITEMDATA, i, 0);
	
	path2 = BrowseFolder(hwnd, path1, IDS_BROWSE_SRC);
	if(path2 != NULL) {
		SendMessage(cb, CB_SETCURSEL, -1, 0);
		SetWindowText(cb, path2);
		free(path2);
	}
	if(i == CB_ERR)
		free(path1);
	
	return TRUE;
}

BOOL OnBrowseDest(HWND hwnd)
{
	char *path1, *path2;
	HWND ed = GetDlgItem(hwnd, IDC_DEST);
	int len = GetWindowTextLength(ed);
	path1 = (char *)_alloca(len + 1);
	GetWindowText(ed, path1, len + 1);
	
	path2 = BrowseFolder(hwnd, path1, IDS_BROWSE_DEST);
	if(path2 != NULL) {
		SetWindowText(ed, path2);
		free(path2);
	}
	return TRUE;
}

static bool dest_editted = false;
BOOL OnSelchangePlugin(HWND hwnd)
{
	HWND cb = GetDlgItem(hwnd, IDC_PLUGIN);
	int sel = (int)SendMessage(cb, CB_GETCURSEL, 0, 0);
	Plugin *p = (Plugin *)SendMessage(cb, CB_GETITEMDATA, sel, 0);
	
	char bnote[128], *pnote, *note;
	size_t blen, plen;
	LoadString(g_hinst, IDS_NOTE, bnote, countof(bnote));
	pnote = p->GetNote();
	blen = strlen(bnote);
	plen = strlen(pnote);
	note = (char *)_alloca(blen + plen + 1);
	memcpy(note, bnote, blen);
	memcpy(note + blen, pnote, plen + 1);
	SetDlgItemText(hwnd, IDC_NOTE, note);
	free(pnote);
	
	EnableDlgItem(hwnd, IDC_SETTING, p->SettingDlg != NULL);
	EnableDlgItem(hwnd, IDC_START, true);
	
	char *path = p->GetDefaultDestDir();
	if(path != NULL && dest_editted) {
		char mesg0[128];
		LoadString(g_hinst, IDS_DESTCHANGE, mesg0, countof(mesg0));
		
		size_t mlen = strlen(mesg0);
		size_t plen = strlen(path);
		char *mesg = (char *)_alloca(mlen + plen + 16);
		wsprintf(mesg, mesg0, path);
		
		int r = MyMessageBox(hwnd, mesg, MB_ICONQUESTION | MB_YESNO);
		if(r == IDYES)
			goto set_dest;
	} else if(!dest_editted) {
	set_dest:
		SetDlgItemText(hwnd, IDC_DEST, path ? path : "");
		dest_editted = false;
	}
	free(path);
	
	return TRUE;
}

BOOL OnChangeDest(HWND hwnd)
{
	dest_editted = true;
	return TRUE;
}

BOOL OnSetting(HWND hwnd)
{
	HWND cb = GetDlgItem(hwnd, IDC_PLUGIN);
	int sel = (int)SendMessage(cb, CB_GETCURSEL, 0, 0);
	if(sel != CB_ERR) {
		Plugin *p = (Plugin *)SendMessage(cb, CB_GETITEMDATA, sel, 0);
		if(p->SettingDlg)
			(*p->SettingDlg)(hwnd);
	}
	return TRUE;
}

BOOL OnStart(HWND hwnd)
{
	HWND cb;
	int sel, len;
	char *source, *dest;
	BOOL OnFinishConvert(HWND hwnd);
	
	cb = GetDlgItem(hwnd, IDC_PLUGIN);
	sel = (int)SendMessage(cb, CB_GETCURSEL, 0, 0);
	if(sel == CB_ERR) {
		MyMessageBox(hwnd, IDS_SEL_PLUGIN, MB_ICONEXCLAMATION | MB_OK);
		SendMessage(hwnd, WM_NEXTDLGCTL, (WPARAM)cb, true);
		return TRUE;
	}
	Plugin *p = (Plugin *)SendMessage(cb, CB_GETITEMDATA, sel, 0);
	
	cb = GetDlgItem(hwnd, IDC_DEST);
	len = GetWindowTextLength(cb);
	if(len == 0) {
		MyMessageBox(hwnd, IDS_NO_DEST, MB_ICONEXCLAMATION | MB_OK);
		SendMessage(hwnd, WM_NEXTDLGCTL, (WPARAM)cb, true);
		return TRUE;
	}
	dest = (char *)malloc(len + 1);
	GetWindowText(cb, dest, len + 1);
	
	cb = GetDlgItem(hwnd, IDC_SOURCE);
	sel = (int)SendMessage(cb, CB_GETCURSEL, 0, 0);
	if(sel == CB_ERR) {
		len = GetWindowTextLength(cb);
		source = (char *)malloc(len + 1);
		GetWindowText(cb, source, len + 1);
	} else {
		char *source0 = (char *)SendMessage(cb, CB_GETITEMDATA, sel, 0);
		source = strdup(source0);
	}
	
	EnableDlgItem(hwnd, IDC_ABORT, true);
	SendMessage(hwnd, WM_NEXTDLGCTL, (WPARAM)GetDlgItem(hwnd, IDC_ABORT), true);
	for(int i = 0; i < countof(disable_ids); ++i)
		EnableDlgItem(hwnd, disable_ids[i], false);
	EnableDlgItem(hwnd, IDC_SETTING, false);
	EnableMenuItem(GetSystemMenu(hwnd, false), SC_CLOSE, MF_GRAYED | MF_BYCOMMAND);
	
	g_pause = g_abort = false;
	if(!StartConvert(hwnd, source, dest, p))
		OnFinishConvert(hwnd);
	
	return TRUE;
}

BOOL OnFinishConvert(HWND hwnd)
{
	HWND cb = GetDlgItem(hwnd, IDC_PLUGIN);
	int sel = (int)SendMessage(cb, CB_GETCURSEL, 0, 0);
	Plugin *p = (Plugin *)SendMessage(cb, CB_GETITEMDATA, sel, 0);
	
	for(int i = 0; i < countof(disable_ids); ++i)
		EnableDlgItem(hwnd, disable_ids[i], true);
	SendMessage(hwnd, WM_NEXTDLGCTL, (WPARAM)GetDlgItem(hwnd, IDC_SOURCE), true);
	EnableDlgItem(hwnd, IDC_ABORT, false);
	EnableDlgItem(hwnd, IDC_SETTING, p->SettingDlg != NULL);
	EnableMenuItem(GetSystemMenu(hwnd, false), SC_CLOSE, MF_ENABLED | MF_BYCOMMAND);
	
	SetDlgItemText(hwnd, IDC_CONV_FOLDER, "");
	SetDlgItemText(hwnd, IDC_FOLDER_RATIO, "");
	SendDlgItemMessage(hwnd, IDC_PROGRESS, PBM_SETPOS, 0, 0);
	
	return TRUE;
}

BOOL OnAbort(HWND hwnd)
{
	g_pause = true;
	if(MyMessageBox(hwnd, IDS_CONFIRM, MB_ICONEXCLAMATION | MB_YESNO | MB_DEFBUTTON2) == IDYES)
		g_abort = true;
	g_pause = false;
	return TRUE;
}

BOOL OnCancel(HWND hwnd)
{
	g_abort = true;
	EndDialog(hwnd, IDCANCEL);
	return TRUE;
}

BOOL CALLBACK DialogProc(HWND hwnd, UINT msg, WPARAM wp, LPARAM lp)
{
	switch(msg) {
	case WM_INITDIALOG:
		return OnInitDialog(hwnd);
	case WM_DESTROY:
		return OnDestroy(hwnd);
	case WM_COMMAND:
		switch(LOWORD(wp)) {
		case IDC_BROWSE_SRC:
			return OnBrowseSrc(hwnd);
		case IDC_BROWSE_DEST:
			return OnBrowseDest(hwnd);
		case IDC_START:
			return OnStart(hwnd);
		case IDC_ABORT:
			return OnAbort(hwnd);
		case IDCANCEL:
			return OnCancel(hwnd);
		case IDC_PLUGIN:
			if(HIWORD(wp) == CBN_SELCHANGE)
				return OnSelchangePlugin(hwnd);
			break;
		case IDC_DEST:
			if(HIWORD(wp) == EN_CHANGE)
				return OnChangeDest(hwnd);
			break;
		case IDC_SETTING:
			return OnSetting(hwnd);
		}
		break;
	case WM_FINISH_CONVERT:
		return OnFinishConvert(hwnd);
	}
	return FALSE;
}

int WINAPI WinMain(HINSTANCE hinst, HINSTANCE, LPSTR, int)
{
	_CrtSetDbgFlag(_CrtSetDbgFlag(_CRTDBG_REPORT_FLAG) | _CRTDBG_LEAK_CHECK_DF);
	
	g_hinst = hinst;
	DialogBox(hinst, MAKEINTRESOURCE(IDD_OE2_DIALOG), NULL, DialogProc);
	return 0;
}
