// OE2eml.cpp - Outlook-Express-To - OE5/6 Multi Converter
//              eml (message/rfc822) writer plugin
// Copyright (C) 2002 Tietew
//
#define WIN32_LEAN_AND_MEAN
#define STRICT
#include <windows.h>
#include <malloc.h>
#include <algorithm>
#pragma hdrstop

#include "resource.h"

#ifndef _DEBUG
#define DllMain _DllMainCRTStartup
#endif

HINSTANCE g_hInstance;

extern "C" BOOL WINAPI DllMain(HINSTANCE module, DWORD reason, LPVOID reserved)
{
	if(reason == DLL_PROCESS_ATTACH) {
		g_hInstance = module;
		DisableThreadLibraryCalls(module);
	}
	return TRUE;
}

extern "C" LPCSTR WINAPI GetPluginName()
{
	return MAKEINTRESOURCE(IDS_PLUGINNAME);
}

extern "C" LPCSTR WINAPI GetNote()
{
	return MAKEINTRESOURCE(IDS_NOTE);
}

BOOL CALLBACK SettingProc(HWND hwnd, UINT msg, WPARAM wp, LPARAM lp);
extern "C" BOOL WINAPI SettingDlg(HWND hwnd)
{
	return DialogBox(g_hInstance, MAKEINTRESOURCE(IDD_SETTING), hwnd, SettingProc);
}

#define SUBJECT_MAX 128
HANDLE g_hfile;
DWORD g_sequence;
int g_type;
char g_subject[SUBJECT_MAX + 10];
enum { sequence, sequence_noext, subject } g_naming_rule;

extern "C" BOOL WINAPI StartConvert(LPCSTR destdir)
{
	return SetCurrentDirectory(destdir);
}

extern "C" BOOL WINAPI NewFolder(LPCSTR folder)
{
	g_sequence = 0;
	CreateDirectory(folder, NULL);
	return SetCurrentDirectory(folder);
}

extern "C" BOOL WINAPI NewMessage(int type)
{
	char fname[MAX_PATH];
	switch(g_naming_rule) {
	case sequence:
		wsprintf(fname, "%08d.%s", ++g_sequence, type == 1 ? "eml" : "nws");
		break;
	case sequence_noext:
		wsprintf(fname, "%d", ++g_sequence);
		break;
	case subject:
		wsprintf(fname, "temp%d.eml", ++g_sequence);
		g_subject[0] = '\0';
		break;
	default:
		__assume(false);
	}
	g_type = type;
	
	g_hfile = CreateFile(fname, GENERIC_WRITE, 0, NULL, CREATE_ALWAYS, FILE_ATTRIBUTE_NORMAL, NULL);
	if(g_hfile == INVALID_HANDLE_VALUE)
		return false;
	return true;
}

extern "C" BOOL WINAPI AppendBuffer(const void *buff, DWORD size)
{
	if(g_naming_rule == subject && g_subject[0] == '\0') {
		static char lastline[SUBJECT_MAX + 10];
		static int lastlen = 0;
		const BYTE *pch = (const BYTE *)buff;
		const BYTE *last = pch + size;
		const BYTE *nl;
		while((nl = std::find(pch, last, '\n')) != last) {
			int len = nl - pch;
			for(int i = -1; nl + i >= pch && nl[i] <= ' '; --i, --len)
				;
			if(len >= SUBJECT_MAX - lastlen)
				len = SUBJECT_MAX - lastlen;
			memcpy(lastline + lastlen, pch, len);
			lastlen += len;
			lastline[lastlen] = '\0';
			
			CharLowerBuff(lastline, 7);
			if(memcmp(lastline, "subject:", 8) == 0) {
				int i;
				for(i = 8; i < lastlen; ++i)
					if(lastline[i] != ' ' && lastline[i] != '\t')
						break;
				memcpy(g_subject, lastline + i, lastlen - i);
				g_subject[lastlen - i] = '\0';
				lastlen = 0;
				goto found;
			}
			pch = nl + 1;
			lastlen = 0;
		}
		if(pch == last)
			lastlen = 0;
		else {
			int len = last - pch;
			if(len >= SUBJECT_MAX - lastlen)
				len = SUBJECT_MAX - lastlen;
			memcpy(lastline, pch, len);
			lastlen = len;
		}
	}
found:;
	
	DWORD written;
	return WriteFile(g_hfile, buff, size, &written, NULL);
}

// flag if the character can be used as filename
const char fnok[] = {
	0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0,
	0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0,
	1, 1, 0, 1, 1, 1, 1, 1, 1, 1, 0, 1, 1, 1, 1, 0,  //  !"#$%&'()*+,-./
	1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 0, 0, 0, 0, 0, 0,  // 0123456789:;<=>?
	1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1,  // @ABCDEFGHIJKLMNO
	1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 0, 1, 1, 1,  // PQRSTUVWXYZ[\]^_
	1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1,  // `abcdefghijklmno
	1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 0, 1, 1, 0,  // pqrstuvwxyz{|}~
};

extern "C" BOOL WINAPI FinishMessage()
{
	if(!CloseHandle(g_hfile))
		return false;
	if(g_naming_rule == subject) {
		char fname[SUBJECT_MAX + 10];
		int len = strlen(g_subject);
		if(len == 0) {
			memcpy(g_subject, "noname", 7);
			len = 6;
		} else {
			for(int i = 0; i < len; ++i)
				if((unsigned char)g_subject[i] > 0x80 || !fnok[g_subject[i]])
					g_subject[i] = '=';
		}
		memcpy(fname, g_subject, len);
		fname[len] = '.';
		memcpy(fname + len + 1, g_type == 1 ? "eml" : "nws", 4);
		
		int i = 1;
		WIN32_FIND_DATA fd;
		HANDLE hfind = FindFirstFile(fname, &fd);
		while(hfind != INVALID_HANDLE_VALUE) {
			FindClose(hfind);
			wsprintf(fname + len, " (%d).%s", i++, g_type == 1 ? "eml" : "nws");
			hfind = FindFirstFile(fname, &fd);
		}
		
		char tempname[32];
		wsprintf(tempname, "temp%d.eml", g_sequence);
		return MoveFile(tempname, fname);
	}
	return true;
}

extern "C" BOOL WINAPI FinishFolder()
{
	return SetCurrentDirectory("..");
}

extern "C" BOOL WINAPI FinishConvert()
{
	return true;
}

BOOL CALLBACK SettingProc(HWND hwnd, UINT msg, WPARAM wp, LPARAM lp)
{
	switch(msg) {
	case WM_INITDIALOG:
		switch(g_naming_rule) {
		case sequence:
			CheckDlgButton(hwnd, IDC_SEQUENCE, BST_CHECKED);
			break;
		
		case sequence_noext:
			CheckDlgButton(hwnd, IDC_SEQUENCE_NOEXT, BST_CHECKED);
			break;
		
		case subject:
			CheckDlgButton(hwnd, IDC_SUBJECT, BST_CHECKED);
			break;
		
		default:
			__assume(false);
		}
		return TRUE;
	
	case WM_COMMAND:
		switch(wp) {
		case IDOK:
			if(IsDlgButtonChecked(hwnd, IDC_SEQUENCE) == BST_CHECKED)
				g_naming_rule = sequence;
			else if(IsDlgButtonChecked(hwnd, IDC_SEQUENCE_NOEXT) == BST_CHECKED)
				g_naming_rule = sequence_noext;
			else
				g_naming_rule = subject;
		
		case IDCANCEL:
			EndDialog(hwnd, wp);
			return TRUE;
		}
		break;
	}
	return FALSE;
}
