// OE2mbox.cpp - Outlook-Express-To - OE5/6 Multi Converter
//               UNIX mbox writer plugin
// Copyright (C) 2002 Tietew
//
#define WIN32_LEAN_AND_MEAN
#define STRICT
#include <windows.h>
#include <malloc.h>
#include <algorithm>
#pragma hdrstop

#include "resource.h"

#ifndef _DEBUG
#define DllMain _DllMainCRTStartup
#endif
HINSTANCE g_hInstance;

extern "C" BOOL WINAPI DllMain(HINSTANCE module, DWORD reason, LPVOID reserved)
{
	if(reason == DLL_PROCESS_ATTACH) {
		g_hInstance = module;
		DisableThreadLibraryCalls(module);
	}
	return TRUE;
}

extern "C" LPCSTR WINAPI GetPluginName()
{
	return MAKEINTRESOURCE(IDS_PLUGINNAME);
}

extern "C" LPCSTR WINAPI GetNote()
{
	return MAKEINTRESOURCE(IDS_NOTE);
}

static BOOL CALLBACK SettingProc(HWND hwnd, UINT msg, WPARAM wp, LPARAM lp);
extern "C" BOOL WINAPI SettingDlg(HWND hwnd)
{
	return DialogBox(g_hInstance, MAKEINTRESOURCE(IDD_SETTING), hwnd, SettingProc);
}

HANDLE g_hfile;
enum { from_quote, ignore_case, pass_through } g_mode = from_quote;
LPSTR g_line;
int g_line_len, g_line_cap;
LPSTR g_header;
int g_header_len, g_header_cap;
bool g_reading_header;
LPSTR g_from;
char g_date[64];

extern "C" BOOL WINAPI StartConvert(LPCSTR destdir)
{
	static const char *const months[] = { "Jan", "Feb", "Mar", "Apr", "May", "Jun",
										  "Jul", "Aug", "Sep", "Oct", "Nov", "Dec" };
	static const char *const wdays[] = { "Sun", "Mon", "Tue", "Wed", "Thu", "Fri", "Sat" };
	
	g_line = (LPSTR)GlobalAlloc(GMEM_FIXED, g_line_cap = 1024);
	g_header = (LPSTR)GlobalAlloc(GMEM_FIXED, g_header_cap = 65536);
	if(g_line == NULL || g_header == NULL) {
		SetLastError(ERROR_NOT_ENOUGH_MEMORY);
		return false;
	}
#ifdef _DEBUG
	memset(g_line, 0, g_line_cap);
	memset(g_header, 0, g_header_cap);
#endif
	
	SYSTEMTIME st;
	GetLocalTime(&st);
	wsprintf(g_date, " %s %s %02d %02d:%02d:%02d %04d\r\n",
		wdays[st.wDayOfWeek], months[st.wMonth - 1], st.wDay,
		st.wHour, st.wMinute, st.wSecond, st.wYear);
	
	return SetCurrentDirectory(destdir);
}

extern "C" BOOL WINAPI NewFolder(LPCSTR folder)
{
	int len = strlen(folder);
	char *fname = (char *)_alloca(len + 5);
	memcpy(fname, folder, len);
	memcpy(fname + len, ".mbx", 5);
	
	g_hfile = CreateFile(fname, GENERIC_WRITE, 0, NULL, CREATE_ALWAYS, FILE_ATTRIBUTE_NORMAL, NULL);
	return g_hfile != INVALID_HANDLE_VALUE;
}

extern "C" BOOL WINAPI AppendBuffer(const void *buff, DWORD size);
extern "C" BOOL WINAPI NewMessage(int type)
{
	g_reading_header = true;
	g_line_len = 0;
	g_header_len = 0;
	return true;
}

static LPSTR ParseAddress()
{
	LPCSTR last = (LPCSTR)(g_line + g_line_len);
	LPCSTR first = std::find((LPCSTR)g_line, last, ':');
	if(first == last)
		return NULL;
	
	for(++first; first < last && *first <= ' '; ++first)
		;
	LPCSTR pch = first;
	while(pch < last) {
		switch(*pch) {
		case '\"':
			for(; pch < last && *pch != '\"'; ++pch)
				if(*pch == '\\')
					++pch;
			break;
		case '(':
			if(first != pch)
				goto found;
			for(; pch < last && *pch != ')'; ++pch)
				if(*pch == '\\')
					++pch;
			break;
		case '<':
			first = pch + 1;
			break;
		case '>':
			goto found;
		}
		++pch;
	}
found:
	for(--pch; pch >= first && *pch <= ' '; --pch)
		;
	++pch;
	
	int len = pch - first;
	LPSTR from = (LPSTR)GlobalAlloc(GMEM_FIXED, len + 1);
	memcpy(from, first, len);
	from[len] = '\0';
	return from;
}

static BOOL WriteBuffer(const void *buff, DWORD size)
{
	DWORD written;
	return WriteFile(g_hfile, buff, size, &written, NULL);
}

static BOOL AppendLine()
{
	if(g_reading_header) {
		if(g_header_len + g_line_len > g_header_cap) {
			LPSTR header2 = (LPSTR)GlobalAlloc(GMEM_FIXED, g_header_cap *= 2);
			if(header2 == NULL) {
				SetLastError(ERROR_NOT_ENOUGH_MEMORY);
				return false;
			}
		#ifdef _DEBUG
			memset(header2, 0, g_header_cap);
		#endif
			memcpy(header2, g_header, g_header_len);
			GlobalFree((HGLOBAL)g_header);
			g_header = header2;
		}
		memcpy(g_header + g_header_len, g_line, g_line_len);
		g_header_len += g_line_len;
		
		if(g_from == NULL) {
			char head[16];
			memcpy(head, g_line, 12);
			CharLowerBuff(head, 12);
			if(memcmp(head, "return-path:", 12) == 0 || memcmp(head, "from:", 5) == 0)
				g_from = ParseAddress();
		}
		if(g_line[0] == '\r' && g_line[1] == '\n')
		{
			if(!WriteBuffer("From ", 5))
				return false;
			if(g_from == NULL) {
				if(!WriteBuffer("unknown@oe2", 11))
					return false;
			} else {
				if(!WriteBuffer(g_from, strlen(g_from)))
					return false;
			}
			if(!WriteBuffer(g_date, strlen(g_date)))
				return false;
			if(!WriteBuffer(g_header, g_header_len))
				return false;
			g_reading_header = false;
		#ifdef _DEBUG
			memset(g_header, 0, g_header_len);
		#endif
		}
		return true;
	}
	if(g_mode != pass_through) {
		int i = 0;
		if(g_line[0] == '>')
			for(i = 1; i < g_line_len && g_line[i] == '>'; ++i)
				;
		if(i + 4 <= g_line_len) {
			bool quote = false;
			char from[5];
			if(g_mode == from_quote)
				quote = (memcmp(g_line, "From ", 5) == 0);
			else {
				memcpy(from, g_line + i, 4);
				CharLowerBuff(from, 4);
				quote = (memcmp(from, "from", 4) == 0);
			}
			if(quote)
				if(!WriteBuffer(">", 1))
					return false;
		}
	}
	return WriteBuffer(g_line, g_line_len);
}

extern "C" BOOL WINAPI AppendBuffer(const void *buff, DWORD size)
{
	const char *first = (const char *)buff;
	const char *last = first + size;
	const char *pch;
	while(1) {
		pch = std::find(first, last, '\n');
		int len = pch - first;
		if(pch != last)
			++len;
		if(g_line_len + len + 1 > g_line_cap) {
			LPSTR line2 = (LPSTR)GlobalAlloc(GMEM_FIXED, g_line_cap *= 2);
			if(line2 == NULL) {
				SetLastError(ERROR_NOT_ENOUGH_MEMORY);
				return false;
			}
		#ifdef _DEBUG
			memset(line2, 0, g_line_cap);
		#endif
			memcpy(line2, g_line, g_line_len);
			GlobalFree((HGLOBAL)g_line);
			g_line = line2;
		}
		memcpy(g_line + g_line_len, first, len);
		g_line_len += len;
		
		if(pch == last)
			break;
		if(!AppendLine())
			return false;
	#ifdef _DEBUG
		memset(g_line, 0, g_line_len);
	#endif
		g_line_len = 0;
		first = pch + 1;
	}
	return true;
}

extern "C" BOOL WINAPI FinishMessage()
{
	if(!WriteBuffer("\r\n", 2))
		return false;
	if(g_from) {
		GlobalFree((HGLOBAL)g_from);
		g_from = NULL;
	}
	return true;
}

extern "C" BOOL WINAPI FinishFolder()
{
	if(!CloseHandle(g_hfile))
		return false;
	g_hfile = NULL;
	return true;
}

extern "C" BOOL WINAPI FinishConvert()
{
	GlobalFree((HGLOBAL)g_line);
	GlobalFree((HGLOBAL)g_header);
	g_line = g_header = NULL;
	
	return true;
}

BOOL CALLBACK SettingProc(HWND hwnd, UINT msg, WPARAM wp, LPARAM lp)
{
	switch(msg) {
	case WM_INITDIALOG:
		switch(g_mode) {
		case from_quote:
			CheckDlgButton(hwnd, IDC_QUOTE, BST_CHECKED);
			break;
		
		case ignore_case:
			CheckDlgButton(hwnd, IDC_IGNORECASE, BST_CHECKED);
			break;
		
		case pass_through:
			CheckDlgButton(hwnd, IDC_PASSTHRU, BST_CHECKED);
			break;
		
		default:
			__assume(false);
		}
		return TRUE;
	
	case WM_COMMAND:
		switch(wp) {
		case IDOK:
			if(IsDlgButtonChecked(hwnd, IDC_QUOTE) == BST_CHECKED)
				g_mode = from_quote;
			else if(IsDlgButtonChecked(hwnd, IDC_IGNORECASE) == BST_CHECKED)
				g_mode = ignore_case;
			else
				g_mode = pass_through;
		
			// fall through
		case IDCANCEL:
			EndDialog(hwnd, wp);
			return TRUE;
		}
		break;
	}
	return FALSE;
}
