// convert.cpp - OutlookExpress-To - OE5/6 Multi Converter
// Copyright (C) 2002 Tietew
//
#include "pch.h"
#include "OE2.h"
#include "utils.h"
#include <vector>
#include <algorithm>

#pragma comment(lib, "winmm.lib")

static int static_pv;

template<> struct std::less<char *> : public std::binary_function<char *, char *, bool> {
	__forceinline bool operator()(const char *lhs, const char *rhs) const {
		return lstrcmpi(lhs, rhs) < 0;
	}
};
struct free_str : public std::unary_function<char *, void> {
	__forceinline void operator()(char *str) {
		free(str);
	}
};

struct ErrorInfo {
	DWORD errcode;
	LPCSTR folder;
	bool dbxerror;
};

BOOL CALLBACK ErrorDlg(HWND hwnd, UINT msg, WPARAM wp, LPARAM lp)
{
	static UINT last_pv = IDC_RETRY;
	
	switch(msg) {
	case WM_INITDIALOG: {
		const ErrorInfo& ctxt = *(const ErrorInfo *)lp;
		
		LPCSTR lmesg;
		FormatMessage(
			FORMAT_MESSAGE_FROM_SYSTEM |
			FORMAT_MESSAGE_ALLOCATE_BUFFER |
			FORMAT_MESSAGE_IGNORE_INSERTS,
			NULL, ctxt.errcode,
			MAKELANGID(LANG_NEUTRAL, SUBLANG_DEFAULT),
			(LPSTR)&lmesg, 0, NULL);
		SetDlgItemText(hwnd, IDC_FOLDER, ctxt.folder);
		SetDlgItemText(hwnd, IDC_ERROR_DETAIL, lmesg);
		LocalFree((HLOCAL)lmesg);
		
		SendDlgItemMessage(hwnd, IDC_HANDICON, STM_SETICON,
			(WPARAM)LoadIcon(NULL, IDI_HAND), 0);
		
		if(ctxt.dbxerror) {
			EnableDlgItem(hwnd, IDC_SKIP_MAIL, false);
			if(last_pv == IDC_SKIP_MAIL)
				last_pv = IDC_RETRY;
		}
		CheckDlgButton(hwnd, last_pv, BST_CHECKED);
		if(last_pv == IDC_RETRY || last_pv == IDC_ABORT)
			EnableDlgItem(hwnd, IDC_ALL_OK, false);
		
		EnableMenuItem(GetSystemMenu(hwnd, false), SC_CLOSE, MF_GRAYED | MF_BYCOMMAND);
		SendMessage(hwnd, WM_NEXTDLGCTL, (WPARAM)GetDlgItem(hwnd, IDOK), true);
		
		PlaySound("SystemHand", NULL, SND_ALIAS | SND_ASYNC | SND_NODEFAULT);
		return FALSE;
	}
	case WM_COMMAND:
		switch(wp) {
		case IDOK:
		case IDC_ALL_OK: {
			static const int spv[] = { IDC_RETRY, IDC_SKIP_MAIL, IDC_SKIP_FOLDER, IDC_ABORT };
			for(int i = 0; i < countof(spv); ++i)
				if(IsDlgButtonChecked(hwnd, spv[i]) == BST_CHECKED) {
					if(wp == IDC_ALL_OK)
						static_pv = spv[i];
					last_pv = spv[i];
					EndDialog(hwnd, spv[i]);
					break;
				}
			return TRUE;
		}
		case IDC_RETRY:
		case IDC_ABORT:
			EnableDlgItem(hwnd, IDC_ALL_OK, false);
			return TRUE;
		
		case IDC_SKIP_MAIL:
		case IDC_SKIP_FOLDER:
			EnableDlgItem(hwnd, IDC_ALL_OK, true);
			return TRUE;
		}
		break;
	}
	return FALSE;
}

int Error(HWND hwnd, LPCSTR folder, bool dbxerror)
{
	int pv = static_pv;
	if((dbxerror && pv == IDC_SKIP_MAIL) || pv == -1) {
		ErrorInfo ctxt;
		ctxt.errcode = GetLastError();
		ctxt.folder = folder;
		ctxt.dbxerror = dbxerror;
		
		pv = (int)DialogBoxParam(g_hinst, MAKEINTRESOURCE(IDD_CONVERT_ERROR),
			hwnd, ErrorDlg, (LPARAM)&ctxt);
	}
	return pv;
}

void GetIndex(FileAccessor& ptr, DWORD table_ptr, std::vector<DWORD>& index)
{
	DWORD another_ptr = ptr.dword(table_ptr + 0x08);
	DWORD ptr_count = ptr.dword(table_ptr + 0x11) & 0x00FFFFFF;
	DWORD index_count = ptr.dword(table_ptr + 0x14);
	if(index_count > 0)
		GetIndex(ptr, another_ptr, index);
	table_ptr += 0x18;
	for(DWORD i = 0; i < ptr_count; ++i) {
		DWORD index_ptr = ptr.dword(table_ptr);
		another_ptr = ptr.dword(table_ptr + 0x04);
		index_count = ptr.dword(table_ptr + 0x08);
		index.push_back(index_ptr);
		table_ptr += 0x0C;
		if(index_count > 0)
			GetIndex(ptr, another_ptr, index);
	}
}

bool ConvertDbx(HWND hwnd, FileAccessor& ptr, Writer& wr, int& count)
{
	std::vector<DWORD> index;
	GetIndex(ptr, ptr.dword(0xE4), index);
	if(index.empty())
		return true;
	
	int i = 0;
	for(std::vector<DWORD>::iterator it = index.begin(); it != index.end(); ++it, ++i) {
		while(g_pause)
			Sleep(10);
		if(g_abort)
			return false;
		
	retry:
		SendDlgItemMessage(hwnd, IDC_PROGRESS, PBM_SETPOS, 100 * i / index.size(), 0);
		
		DWORD news_flag = ptr.dword(*it + 0x1C);
		DWORD mail_flag = ptr.dword(*it + 0x18);
		int type = -1;
		
		DWORD data_ptr;
		if((news_flag & 0x0F) == 4) {
			type = 2;
			data_ptr = news_flag;
		} else if((mail_flag & 0x0F) == 4) {
			type = 1;
			data_ptr = mail_flag;
		}
		if(type == -1)
			continue;
		
		if((data_ptr & 0xFF) >= 0x80)
			data_ptr >>= 8;
		else {
			BYTE b1 = (BYTE)(data_ptr >> 8);
			BYTE b2 = (BYTE)(ptr.dword(*it + 0x0A));
			data_ptr = ptr.dword(*it + 0x0C + b2 * 4 + b1);
		}
		
		bool last_lf = true;
		if(!wr.NewMessage(type))
			goto error;
		while(data_ptr) {
			DWORD data_len = ptr.dword(data_ptr + 0x08);
			DWORD next_ptr = ptr.dword(data_ptr + 0x0C);
			
			BYTE buffer[2048];
			DWORD copy_len;
			while(data_len != 0) {
				copy_len = std::min<DWORD>(sizeof(buffer), data_len);
				if(ptr.buffer(buffer, data_ptr + 0x10, copy_len)) {
					if(!wr.AppendBuffer(buffer, data_len))
						goto error;
					last_lf = buffer[copy_len - 1] == '\n';
					data_ptr += copy_len;
					data_len -= copy_len;
				}
			}
			data_ptr = next_ptr;
		}
		if(!last_lf && !wr.AppendBuffer("\r\n", 2))
			goto error;
		
		if(!wr.FinishMessage())
			goto error;
		
		++count;
		continue;
	
	error:
		switch(Error(hwnd, wr.Folder(), false)) {
		case IDC_RETRY:
			goto retry;
		case IDC_SKIP_MAIL:
			break;
		case IDC_SKIP_FOLDER:
			return true;
		case IDC_ABORT:
			return false;
		default:
			__assume(false);
		}
	}
	return true;
}

bool ConvertFolder(HWND hwnd, char *dbx, Writer& wr, int& count)
{
retry:
	bool result = false;
	UINT pv = -1;
	UINT errid = IDS_FAIL_OPEN;
	HANDLE fh = INVALID_HANDLE_VALUE;
	
	fh = CreateFile(dbx, GENERIC_READ, FILE_SHARE_READ, NULL,
		OPEN_EXISTING, FILE_ATTRIBUTE_NORMAL, NULL);
	if(fh != INVALID_HANDLE_VALUE) {
		FileAccessor acc(fh);
		result = true;
		if(acc.dword(0xC4) != 0)
			result = ConvertDbx(hwnd, acc, wr, count);
		pv = 0;
	}
	if(pv == -1)
		pv = Error(hwnd, wr.Folder(), true);
	
	switch(pv) {
	case IDC_RETRY:
		goto retry;
	case IDC_SKIP_FOLDER:
		return true;
	case IDC_ABORT:
		return false;
	}
	
	return result;
}

bool CompleteDirectory(char *dir)
{
	WIN32_FIND_DATA fd;
	HANDLE fh;
	char *pd, lastch;
	int skipbs;
	
	if(*(WORD*)dir == MAKEWORD('\\', '\\') || *(WORD*)dir == MAKEWORD('/', '/'))
		skipbs = 4;
	else if(('A' <= dir[0] && dir[0] <= 'Z' || 'a' <= dir[0] && dir[0] <= 'z')
		&& dir[1] == ':' && (dir[2] == '\\' || dir[2] == '/'))
		skipbs = 1;
	else {
		SetLastError(ERROR_BAD_PATHNAME);
		return false;
	}
	
	for(pd = dir; ; pd = CharNext(pd)) {
		if(*pd == '\\' || *pd == '/' || *pd == '\0') {
			if(*pd == '\0' && lastch == '\\') {
				*(pd - 1) = '\0';
				return true;
			}
			char temp = *pd;
			*pd = '\0';
			
			if(skipbs > 0)
				--skipbs;
			else {
				fh = FindFirstFile(dir, &fd);
				if(fh != INVALID_HANDLE_VALUE) {
					FindClose(fh);
					if(!(fd.dwFileAttributes & FILE_ATTRIBUTE_DIRECTORY)) {
						SetLastError(ERROR_ALREADY_EXISTS);
						return false;
					}
				} else if(!CreateDirectory(dir, NULL))
					return false;
			}
			if(temp == '\0')
				return true;
			*pd = '\\';
		}
		lastch = *pd;
	}
}

void DoConvert(HWND hwnd, char *source0, char *dest, Plugin *p)
{
	size_t slen = strlen(source0);
	char *source = (char *)_alloca(slen + MAX_PATH + 1);
	memcpy(source, source0, slen);
	for(char *s = source, *nexts; ; s = nexts) {
		nexts = CharNext(s);
		if(*nexts == '\0') {
			if(*s != '\\' && *s != '/')
				source[slen++] = '\\';
			break;
		}
	}
	memcpy(source + slen, "*.dbx", 6);
	
	std::vector<char *> vdbx;
	WIN32_FIND_DATA fd;
	HANDLE fh = FindFirstFile(source, &fd);
	if(fh != INVALID_HANDLE_VALUE) {
		do {
			static const char *const ignores[] = {
				"Folders", "Offline", "Pop3uidl",
			};
			if(fd.dwFileAttributes & FILE_ATTRIBUTE_DIRECTORY)
				continue;
			fd.cFileName[strlen(fd.cFileName) - 4] = '\0';
			for(int i = 0; i < countof(ignores); ++i)
				if(lstrcmpi(fd.cFileName, ignores[i]) == 0)
					goto cont;
			vdbx.push_back(strdup(fd.cFileName));
		cont:;
		} while(FindNextFile(fh, &fd));
		FindClose(fh);
	}
	if(vdbx.empty()) {
		MyMessageBox(hwnd, IDS_NO_DBX, MB_ICONEXCLAMATION | MB_OK);
		return;
	}
	std::sort(vdbx.begin(), vdbx.end());
	
	if(!CompleteDirectory(dest)) {
		ErrorMessage(hwnd, IDS_FAIL_MKDIR, GetLastError(), MB_ICONHAND);
		goto exit;
	}
	
	static_pv = -1;
	if(!p->StartConvert(dest)) {
		ErrorMessage(hwnd, IDS_FAIL_CONVERT, GetLastError(), MB_ICONHAND | MB_OK);
		goto exit;
	}
	
	int i = 1, count = 0;
	HWND pbar = GetDlgItem(hwnd, IDC_PROGRESS);
	for(std::vector<char *>::iterator it = vdbx.begin(); it != vdbx.end(); ++it) {
		char *fname = *it;
		
		size_t flen = strlen(fname);
		memcpy(source + slen, fname, flen + 1);
		memcpy(source + slen + flen, ".dbx", 5);
		
		char ratio[64];
		wsprintf(ratio, "%d / %d", i++, vdbx.size());
		SetDlgItemText(hwnd, IDC_CONV_FOLDER, fname);
		SetDlgItemText(hwnd, IDC_FOLDER_RATIO, ratio);
		SendMessage(pbar, PBM_SETPOS, 0, 0);
		
		Writer wr(p, fname);
		if(!ConvertFolder(hwnd, source, wr, count)) {
			p->FinishConvert();
			goto exit;
		}
		SendMessage(pbar, PBM_SETPOS, 100, 0);
	}
	p->FinishConvert();
	
	char mesg0[128], mesg[128];
	LoadString(g_hinst, IDS_CONVERTED, mesg0, countof(mesg0));
	wsprintf(mesg, mesg0, count);
	MyMessageBox(hwnd, mesg, MB_ICONINFORMATION | MB_OK);

exit:
	std::for_each(vdbx.begin(), vdbx.end(), free_str());
}

struct ConvertData {
	HWND hwnd;
	char *source;
	char *dest;
	Plugin *p;
};

unsigned int CALLBACK ConvertThread0(LPVOID context)
{
	ConvertData *cd = (ConvertData *)context;
	
	char path[MAX_PATH];
	GetCurrentDirectory(countof(path), path);
	
	DoConvert(cd->hwnd, cd->source, cd->dest, cd->p);
	PostMessage(cd->hwnd, WM_FINISH_CONVERT, 0, 0);
	
	SetCurrentDirectory(path);
	
	free(cd->source);
	free(cd->dest);
	delete cd;
	
	return 0;
}

bool StartConvert(HWND hwnd, char *source, char *dest, Plugin *p)
{
	ConvertData *cd = new ConvertData;
	cd->hwnd = hwnd;
	cd->source = source;
	cd->dest = dest;
	cd->p = p;
	
	DWORD tid;
	HANDLE th = CreateThread(NULL, 0, (LPTHREAD_START_ROUTINE)ConvertThread0, cd, 0, &tid);
	if(th == NULL) {
		free(cd->source);
		free(cd->dest);
		delete cd;
		
		ErrorMessage(hwnd, IDS_FAIL_THREAD, GetLastError(), MB_ICONHAND | MB_OK);
		return false;
	}
	CloseHandle(th);
	return true;
}
